<?php
declare(strict_types=1);

namespace Drupal\generic_templates\EventSubscriber;

use DirectoryIterator;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\path_alias\AliasManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Asset\LibraryDiscoveryInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\core_event_dispatcher\FormHookEvents;
use Drupal\core_event_dispatcher\ThemeHookEvents;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\core_event_dispatcher\EntityHookEvents;
use Drupal\core_event_dispatcher\Event\Theme\ThemeEvent;
use Drupal\core_event_dispatcher\Event\Form\FormAlterEvent;
use Drupal\core_event_dispatcher\Event\Entity\EntityViewEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Drupal\core_event_dispatcher\Event\Theme\ThemeSuggestionsAlterEvent;

/**
 * Generic Templates event subscriber.
 */
class GenericTemplatesSubscriber implements EventSubscriberInterface
{
  protected const PREFIX = '__generic';
  protected array $entity_type_ids = [];

  protected array $libraries = [];
  private const ASSET_TYPES = [
    'css' => null,
    'js' => null
  ];

  public function __construct(
    private readonly ConfigFactoryInterface $configFactory,
    private readonly ModuleExtensionList $moduleExtensionList,
    private readonly FileSystemInterface $fileSystem,
    private readonly LibraryDiscoveryInterface $libraryDiscovery,
    private readonly ThemeManagerInterface $themeManager,
    private readonly ThemeHandlerInterface $themeHandler,
    private readonly LanguageManagerInterface $languageManager,
    private readonly AliasManagerInterface $pathAliasManager,
  ) {
    $this->entity_type_ids = $this->configFactory->get('generic_templates.settings')->get('types') ?: [];
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array
  {
    return [
        // @see https://www.drupal.org/docs/7/creating-custom-modules/howtos/using-template-files-in-your-own-module#drupal8
      ThemeHookEvents::THEME => 'theme',
      ThemeHookEvents::THEME_SUGGESTIONS_ALTER => 'themeSuggestionsAlter',

        // Add generic classes to entity attributes
      EntityHookEvents::ENTITY_VIEW => 'entityView',
      FormHookEvents::FORM_ALTER => 'formAlter'
    ];
  }

  /**
   * Add generic themes {entity_type}--generic to suggestions
   *
   * @param ThemeSuggestionsAlterEvent $event
   * @return void
   */
  public function themeSuggestionsAlter(ThemeSuggestionsAlterEvent $event): void
  {
    $hook = $event->getHook();

    if (in_array($hook, $this->entity_type_ids)) {
      // use the generic template lastly by adding it to beginning using array_unshift
      array_unshift($event->getSuggestions(), $hook . self::PREFIX);
    }
  }

  /**
   * Register templates located in generic_templates/templates
   *
   * @see generic_templates/templates folder
   *
   * @param ThemeEvent $event
   * @return void
   */
  public function theme(ThemeEvent $event)
  {
    $path = $this->moduleExtensionList->getPath('generic_templates') . '/templates/entity';

    foreach ($this->entity_type_ids as $id) {
      $template_name = $id . self::PREFIX;
      $file_name = $path . '/' . str_replace('_', '-', $template_name) . '.html.twig';
      if (file_exists($file_name)) {
        $event->addNewTheme($template_name, ['path' => $path, 'base hook' => $id]);
      }
    }
  }

  /**
   * trims entity_type_id prefixes for popular contrib modules
   * @param string $type_id
   * @return string
   */
  private function minimalEntityTypeId(string $type_id)
  {
    $prefix_to_remove = [
      "commerce_",
      "crm_core_"
    ];

    foreach ($prefix_to_remove as $prefix) {
      if (strpos($type_id, $prefix) === 0) {
        $type_id = substr($type_id, strlen($prefix));
        break;
      }
    }
    return $type_id;
  }

  protected function classify(array &$attributes, ContentEntityInterface $entity, string $view_mode): void
  {
    $type_id = $entity->getEntityTypeId();
    $type_id = $this->minimalEntityTypeId($type_id);

    $attributes = NestedArray::mergeDeep($attributes, [
      'class' => [
        'generic',
        Html::getClass($type_id),
        Html::getClass($entity->bundle()),
        Html::getClass($view_mode),
      ]
    ]);
  }

  // Idea: Store a basefield "library_name" in Segment and use this instead
  private function libraryNameFromPath($entity): string
  {
    $url = $entity->toUrl('canonical', [
      'language' => $this->languageManager->getDefaultLanguage()
    ]);
    $alias = explode('/', $this->pathAliasManager->getAliasByPath($url->toString()));

    // Simple file-schema, may be improved
    return implode('.', [
      $entity->getEntityTypeId(),
      end($alias)
    ]);
  }

  /**
   * @todo User provided library patterns at route
   * generic_templates.settings_form using tokens
   * 
   * examples:
   * "[entity:type].[entity:bundle]'
   * "[entity:type].[entity:bundle].[view_mode]'
   * "[entity:type].[entity:bundle].[view_mode].[entity:id]'
   */
  private function getLibrarySuggestions(ContentEntityInterface $entity, string $view_mode)
  {
    // For now, the "patterns" are hardcoded:
    $suggestions = [];
    if ($entity->getEntityTypeId() == 'segment') {
      $suggestions[] = $this->libraryNameFromPath($entity);
    } else {
      $suggestions[] = $entity->getEntityTypeId();
      $suggestions[] = implode('.', [
        $entity->getEntityTypeId(),
        $entity->bundle()
      ]);
      $suggestions[] = implode('.', [
        $entity->getEntityTypeId(),
        $view_mode
      ]);
      $suggestions[] = implode('.', [
        $entity->getEntityTypeId(),
        $entity->bundle(),
        $view_mode
      ]);
    }

    return $suggestions;
  }

  public function entityView(EntityViewEvent $event): void
  {
    $entity = $event->getEntity();

    if (!$entity instanceof ContentEntityInterface)
      return;

    if (in_array($entity->getEntityTypeId(), $this->entity_type_ids)) {
      $build = &$event->getBuild();
      // # Classify
      $build['#attributes'] ??= [];
      $this->classify($build['#attributes'], $entity, $event->getViewMode());

      // # Attach libraries
      $library_suggestions = $this->getLibrarySuggestions($entity, $event->getViewMode());
      foreach ($library_suggestions as $library) {
        if ($this->libraryDiscovery->getLibraryByName('generic_templates', $library)) {
          $build['#attached']['library'][] = "generic_templates/{$library}";
        }
      }
    }
  }

  public function formAlter(FormAlterEvent $event): void
  {
    $form_object = $event->getFormState()->getFormObject();

    if ($form_object instanceof ContentEntityForm) {
      $entity = $form_object->getEntity();

      if (in_array($entity->getEntityTypeId(), $this->entity_type_ids)) {
        // @todo 'full' is not correct... we may call this 'form'?? instead?
        $this->classify($event->getForm()['#attributes'], $entity, 'full');
      }
    }
  }

  private function resolveLibraries(array $current, string $dependency = null, array $stack = [])
  {
    foreach($current as $key => $data) {
      $library = array_intersect_key($data, self::ASSET_TYPES);
      if(!empty($library)) {
        $library_name = implode('.', array_merge($stack, [$key]));
        $this->libraries[$library_name] = $library;
        if($dependency) {
          $this->libraries[$library_name]['dependencies'] = [$dependency];
        }
      }

      if($nested = array_diff_key($data, self::ASSET_TYPES)) {
        $this->resolveLibraries(
          $nested, 
          isset($library_name) ? 'generic_templates/' . $library_name : null, 
          array_merge($stack, [$key])
        );
      }
      unset($library_name);
    }
  }

  public function libraryInfoBuild(): array
  {
    // Only the default theme is scanned for now
    $theme_name = $this->configFactory->get('system.theme')->get('default');
    $theme = $this->themeHandler->getTheme($theme_name);

    $tree = [];
    foreach (['css', 'js'] as $type) {
      $path = $theme->getPath() . '/' . $type . '/generic_templates';

      if (!$this->fileSystem->prepareDirectory($path, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
        \Drupal::logger('generic_templates')->error('Could not create directory: %dir', ['%dir' => $path]);
        return $this->libraries;
      }

      foreach (new DirectoryIterator($path) as $file) {
        // Exclude '.' and '..' directories
        if (!$file->isFile()) {
          continue;
        }
        
        // Climb to tree position
        $parts = explode('.', $file->getBasename('.' . $type));
        $current = &$tree;
        foreach ($parts as $part) {
          if (!isset($current[$part])) {
            $current[$part] = [];
          }
          $current = &$current[$part];
        }

        $current = NestedArray::mergeDeep($current, [
          $type => [
            'theme' => [
              '/' . $file->getPathName() => []
            ]
          ]
        ]);
      }
    }
    
    $this->resolveLibraries($tree);
    return $this->libraries;
  }
}
